from pymusicterm.api.ytmusic import SongData
from pymusicterm.player.player import PyMusicTermPlayer
from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.widgets import (
    Input,
    Button,
    OptionList,
    TabbedContent,
    Label,
    TabPane,
    Rule,
    Select,
    Footer,
    MarkdownViewer,
    ProgressBar,
    Switch,
)
from textual.widgets.option_list import Option
from textual.containers import Horizontal, Vertical
from textual import on
from textual.reactive import reactive
from pathlib import Path
from datetime import timedelta
from textual import work
from textual.worker import get_current_worker
from loguru import logger
import sys
if sys.platform == "win32": 
    from pymusicterm.player.media_control import MediaControlWin32 as MediaControl
else:
    from pymusicterm.player.media_control import MediaControlMPRIS as MediaControl
from pymusicterm.setting import SettingManager, rename_console


def format_time(time: float) -> str:
    """Format the time to a string"""
    return str(timedelta(seconds=int(time))).removeprefix("0:")

class PyMusicTerm(App):
    BINDINGS = [
        ("q", "seek_back", "Seek backward"),
        ("s", "play", "Play/Pause"),
        ("d", "seek_forward", "Seek forward"),
        ("r", "shuffle", "Shuffle"),
        ("l", "loop", "Loop at the end"),
        ("&", "return_on_search_tab", "Go to the search tab"),
        ("é", "return_on_playlist_tab", "Go to the playlist tab"),
        ("k", "volume(0.1)", "Volume up"),
        ("j", "volume(-0.1)", "Volume down"),
    ]

    def __init__(self) -> None:
        super().__init__(css_path="pymusicterm.tcss")
        
        self.setting = SettingManager()
        rename_console("PyMusicTerm")
        logger.remove()
        logger.add(
            Path(self.setting.log_dir + "/pymusicterm.log"),
            format="{time} {level} {message}",
            level="INFO",
        )
        self.timer: Widget | None = None
        self.media_control = MediaControl()
        self.player = PyMusicTermPlayer(self.setting, self.media_control)
        self.media_control.init(self.player)
        
    def compose(self) -> ComposeResult:
        yield Footer()
        with TabbedContent(classes="search_tabs", id="tabbed_content"):
            with TabPane("Search", id="search"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="search_input")
                    yield Select(
                        id="search_sort",
                        options=[
                            ("Songs (default: Stream Youtube Music)", "songs"),
                            ("Video (Stream Youtube Video)", "videos"),
                        ],
                        allow_blank=False,
                    )
                    yield OptionList(id="search_results")
            with TabPane("Playlist", id="playlist"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="playlist_input")
                    yield OptionList(id="playlist_results")
            with TabPane("Lyrics", id="lyrics"):
                with Vertical():
                    yield Input(placeholder="Search for a song", id="lyrics_input")
                    yield MarkdownViewer(
                        markdown="No lyrics now",
                        id="lyrics_results",
                        show_table_of_contents=False,
                    )
        yield Rule()
        with Horizontal(classes="status_controls"):
            yield Label(
                "--:--", id="label_current_song_position", classes="control_label"
            )
            yield ProgressBar(
                total=100, show_eta=False, show_percentage=False, id="player_status"
            )
            yield Label("--:--", id="label_song_length", classes="control_label")
        with Horizontal(classes="player_controls"):
            yield Button("Previous", id="previous")
            yield Button("Play", id="play_pause")
            yield Button("Next", id="next")
        with Horizontal(classes="player_controls"):
            yield Button("Shuffle", id="shuffle")
            yield Button("Loop", id="loop")

    @on(TabbedContent.TabActivated)
    def action_select_playlist_tab(self, event: TabbedContent.TabActivated) -> None:
        """Action when the tab is activated. If the tab is playlist, clear the options and add the songs to the playlist.

        Args:
            event (TabbedContent.TabActivated): The event that triggered the action
        """
        playlist_results: OptionList = self.query_one("#playlist_results")
        if event.tab.id.endswith("playlist"):
            playlist_results.clear_options()
            for i, song in enumerate(self.player.list_of_downloaded_songs):
                playlist_results.add_option(
                    Option(
                        f"{i + 1}. {Path(song).stem}",
                        id=i,
                    )
                )
            return

    def update_time(self) -> None:
        """Update the time label of the player, and update the player"""
        progress_bar: ProgressBar = self.query_one("#player_status")
        label_current_song_position: Label = self.query_one(
            "#label_current_song_position"
        )
        label_song_length: Label = self.query_one("#label_song_length")
        length_float = self.player.song_length
        current_float = self.player.position
        label_current_song_position.update(format_time(current_float))
        label_song_length.update(format_time(length_float))
        percentage = current_float / length_float
        progress_bar.update(
            progress=percentage * 100,
        )
        self.player.update()
        lyrics_results: MarkdownViewer = self.query_one("#lyrics_results")
        if not self.player.playing:
            return
        current_lyric = self.player.dict_of_lyrics[
                self.player.list_of_downloaded_songs[self.player.current_song_index]
            ]
        logger.warning(current_lyric)
        logger.warning(self.player.dict_of_lyrics)
        logger.warning(self.player.list_of_downloaded_songs[self.player.current_song_index])
        with open(current_lyric, "r", encoding="utf-8") as f:
            lyrics_results.document.update(f.read())
            
    def action_return_on_search_tab(self):
        """Set the search tab as the active tab"""
        tab: TabbedContent = self.query_one("#tabbed_content")
        tab.active = "search"

    def action_return_on_playlist_tab(self):
        """Set the playlist tab as the active tab"""
        tab: TabbedContent = self.query_one("#tabbed_content")
        tab.active = "playlist"

    @on(OptionList.OptionSelected, "#playlist_results")
    def select_playlist_result(self, event: OptionList.OptionSelected) -> None:
        """Select a song from the playlist results and play it"""
        id = int(event.option.id)
        self.player.play_from_list(id)
        self.toggle_button()

    @on(OptionList.OptionSelected, "#search_results")
    def select_result(self, event: OptionList.OptionSelected) -> None:
        """Select a song from the search results and play it"""
        video_id = str(event.option.id)
        self.player.play_from_ytb(video_id)
        self.toggle_button()

    @on(Button.Pressed, "#previous")
    def action_previous(self) -> None:
        """Play the previous song"""
        self.player.previous()

    @on(Button.Pressed, "#play_pause")
    def action_play(self) -> None:
        """Play or pause the song"""
        if self.player.playing:
            self.player.pause_song()
        else:
            self.player.resume_song()
        self.toggle_button()

    def toggle_button(self) -> None:
        """Toggle the play button label and start the timer if it's not running"""
        button: Button = self.query_one("#play_pause")
        if self.player.playing:
            button.label = "Pause"
        else:
            button.label = "Play"
        if self.timer is None:
            self.timer = self.set_interval(1, self.update_time, name="update_time")

    @on(Button.Pressed, "#next")
    def action_next(self) -> None:
        """Play the next song"""
        self.player.next()

    @on(Button.Pressed, "#shuffle")
    def action_shuffle(self) -> None:
        """Shuffle the list of downloaded songs"""
        self.player.suffle()

    @on(Button.Pressed, "#loop")
    def action_loop(self) -> None:
        """Toggle the loop button"""
        loop_button: Button = self.query_one("#loop")
        is_looping = self.player.loop_at_end()
        if is_looping:
            loop_button.variant = "success"
        else:
            loop_button.variant = "default"

    @on(Input.Submitted, "#search_input")
    def search(self) -> None:
        """Search for a song on YTMusic and display the results"""
        search_input: Input = self.query_one("#search_input")
        if search_input.value == "":
            return
        search_input.loading = True
        self.search_ytb_thread(search_input.value)

    @work(exclusive=True, thread=True)
    def search_ytb_thread(self, query: str) -> None:
        """Start a thread to search for a song on YTMusic

        Args:
            query (str): The search query
        """
        worker = get_current_worker()
        filter: Select = self.query_one("#search_sort")
        results = self.player.query(query, filter.value)
        if not worker.is_cancelled:
            self.call_from_thread(self.update_search_results, results)

    def update_search_results(self, results: list[SongData]) -> None:
        """Update the search results

        Args:
            results (list): The list of results from youtube music
        """
        search_results: OptionList = self.query_one("#search_results")
        search_input: Input = self.query_one("#search_input")
        search_results.clear_options()
        for i, result in enumerate(results):
            search_results.add_option(
                Option(
                    f"{i + 1}. {result.title} - {result.get_formatted_artists()}",
                    result.videoId,
                )
            )
        search_input.loading = False

    def action_seek_back(self) -> None:
        """Seek backward 10 seconds"""
        self.player.seek(-10)

    def action_seek_forward(self) -> None:
        """Seek forward 10 seconds"""
        self.player.seek(10)

    def action_volume(self, value: float) -> None:
        """Increase the volume"""
        self.player.volume(value)




def main():
    app = PyMusicTerm()
    app.run()


if __name__ == "__main__":
    main()



import ytmusicapi
from ytmusicapi.exceptions import YTMusicError
from dataclasses import dataclass
from typing import Protocol
import ytmusicapi.exceptions
import ytmusicapi.ytmusic
from pymusicterm.api.lyrics import LyricsDownloader

class SongData(Protocol):
    title: str
    duration: str
    videoId: str

    def get_formatted_artists(self) -> str:
        """Format the list of artists to a string

        Returns:
            str: a string with the formatted artists
        """


@dataclass
class SongArtist:
    name: str
    id: str


@dataclass
class SearchResult(SongData):
    title: str
    artist: list[SongArtist]
    duration: int
    videoId: str

    def get_formatted_artists(self) -> str:
        """Get the formatted artists of the song"""
        return ", ".join([artist.name for artist in self.artist])


@dataclass
class LyricsResult:
    lyrics: str
    source: str


class YTMusic:
    def __init__(self, lyrics_downloader: LyricsDownloader):
        self.client = ytmusicapi.YTMusic()
        self.lyrics_downloader = lyrics_downloader

    def search(self, query: str, filter: str = "songs") -> list[SearchResult]:
        """Search for a song on YTMusic
        Args:
            query (str): The query to search for
            filter (str, optional): The filter to use. Defaults to "songs".
        Returns:
            list[Song]: The list of songs found
        """
        results = self.client.search(query, filter)
        return [
            SearchResult(
                title=result["title"],
                artist=[SongArtist(**artist) for artist in result["artists"]],
                duration=result["duration"],
                videoId=result["videoId"],
            )
            for result in results
        ]

    def get_lyrics(self, song: SearchResult) -> LyricsResult | None:
        """Get the lyrics of a song
        Args:
            video_id (str): The video id of the song
        Returns:
            str: The lyrics of the song
        """
        lyrics_id = self.client.get_watch_playlist(song.videoId)["lyrics"]
        try:
            lyrics = self.client.get_lyrics(lyrics_id)
        except ytmusicapi.exceptions.YTMusicError:
            lyrics = None

        if lyrics:
            return self.lyrics_downloader.save(
                LyricsResult(
                    lyrics=lyrics["lyrics"],
                    source=lyrics["source"],
                ),
                song,
            )
        else:
            return self.lyrics_downloader.save(
                LyricsResult(
                    lyrics="None",
                    source="None",
                ),
                song,
            )



from pathlib import Path

from typing import List
from mpris_server.adapters import Metadata, PlayState, MprisAdapter
from mpris_server.base import URI, MIME_TYPES, BEGINNING, DEFAULT_RATE, DbusObj
from mpris_server.server import Server
import sys

class HAdapter(MprisAdapter):
    def __init__(self):
        super().__init__()
        self.player = None
    
    def setup(self, player):
        self.player = player
    
    def get_uri_schemes(self) -> List[str]:
        return URI

    def get_mime_types(self) -> List[str]:
        return MIME_TYPES

    def can_quit(self) -> bool:
        return True

    def quit(self):
        sys.exit()

    def get_current_position(self):
        return self.player.position

    def next(self):
        self.player.next()

    def previous(self):
        self.player.previous()

    def pause(self):
        self.player.pause_song()

    def resume(self):
        self.player.resume_song()

    def stop(self):
        self.player.stop()

    def play(self):
        self.player.playing
        
    def get_playstate(self) -> PlayState:
        if not self.player.playing:
            return PlayState.PAUSED
        else:
            return PlayState.PLAYING

    def seek(self, time):
        self.player.seek_forward(time)

    def is_repeating(self) -> bool:
        return False

    def is_playlist(self) -> bool:
        return self.can_go_next() or self.can_go_previous()

    def set_repeating(self, val: bool):
        self.player.loop_at_end()

    def set_loop_status(self, val: str):
        pass

    def get_rate(self) -> float:
        return 1.0

    def set_rate(self, val: float):
        pass

    def get_shuffle(self) -> bool:
        return False

    def set_shuffle(self, val: bool):
        return False

    def get_art_url(self, track):
        print('Later')
        return 'Later'

    def get_stream_title(self):
        print('Later again')

    def is_mute(self) -> bool:
        return False

    def can_go_next(self) -> bool:
        return True

    def can_go_previous(self) -> bool:
        return  True

    def can_play(self) -> bool:
        return True

    def can_pause(self) -> bool:
        return True

    def can_seek(self) -> bool:
        return True

    def can_control(self) -> bool:
        return True

    def get_stream_title(self) -> str:
        return "Test title"

    def metadata(self) -> dict:
        song_data = Path(self.player.list_of_downloaded_songs[self.player.current_song_index]).stem.split(" - ")
        title = ", ".join(song_data[:-1])
        artist = song_data[-1]
        metadata = {
        "mpris:trackid": "/track/1",
        "mpris:length": self.player.song_length if not 0 else None,
        "mpris:artUrl": "Example",
        "xesam:url": "https://google.com",
        "xesam:title": title,
        "xesam:artist": [artist],
        "xesam:album": "",
        "xesam:albumArtist": [],
        "xesam:discNumber": 1,
        "xesam:trackNumber": 1,
        "xesam:comment": [],
        }

        return metadata
        





from typing import Protocol
from dataclasses import dataclass
from pathlib import Path


@dataclass
class LyricsData(Protocol):
    lyrics: str
    source: str


@dataclass
class SongData(Protocol):
    title: str
    duration: str
    videoId: str

    def get_formatted_artists(self) -> str: ...


class LyricsDownloader:
    def __init__(self, download_path: str):
        self.download_path = download_path

    def save(self, lyrics: LyricsData, song: SongData) -> Path | None:
        """Save the lyrics to a markdown file

        Args:
            lyrics (str): The lyrics to save

        Returns:
            str | None: The path of the saved file or None if the save failed
        """
        lyrics_path = Path(
            f"{self.download_path}/{song.title} - {song.get_formatted_artists()}.md"
        )
        with open(lyrics_path, "w", encoding="utf-8") as f:
            f.write(lyrics.lyrics)
        return lyrics_path



import sys
from typing import Protocol


if sys.platform == "win32": 
    pass
else:
    from mpris_server import EventAdapter
    from mpris_server import EventAdapter, Server
    from pymusicterm.api.mpris.mpris import HAdapter
    
    
class MediaControl(Protocol):
    def init(self) -> None: ...
    def on_playback(self) -> None: ...
    def on_playpause(self) -> None: ...
    def on_volume(self) -> None: ...
    
class MediaControlWin32(MediaControl):
    def __init__(self) -> None:
        print('Win32')

    def init(self, player) -> None:
        print('Win32')
        
    def on_playback(self) -> None:
        print('Win32')
        
    def on_playpause(self) -> None:
        print('Win32')
        
    def on_volume(self) -> None:
        print('Win32')
    
class MediaControlMPRIS(MediaControl):
    def __init__(self) -> None:
        self.adapter = HAdapter()
        self.mpris = Server('PyMusicTerm', adapter=self.adapter)
        self.event = EventAdapter(root=self.mpris.root, player=self.mpris.player)
    
    def init(self, player):
        self.adapter.setup(player)
        self.mpris.loop(background=True)
        
    def on_playback(self):
        return self.event.on_playback()
    
    def on_playpause(self):
        return self.event.on_playpause()
    
    def on_volume(self):
        return self.event.on_volume()


from typing import Protocol
from pytubefix import YouTube
from pydub import AudioSegment
from pathlib import Path


class SongData(Protocol):
    title: str
    duration: str
    videoId: str

    def get_formatted_artists(self) -> str: ...


class Downloader:
    def __init__(self, download_path: str):
        self.download_path = download_path

    def download(self, song: SongData) -> str | None:
        """Download a song from a song object and return the path of the downloaded file.
        If the file already exists, it will not be downloaded again and the path will be returned.

        Args:
            song (Song): The song to download

        Returns:
            path (str): The path of the downloaded file or None if the download failed
        """

        song_path = Path(
            f"{self.download_path}/{song.title} - {song.get_formatted_artists()}.mp3"
        )
        if song_path.exists():
            return str(song_path)

        yt_path = self._download_from_yt(song)
        converted_path = self._convert_to_mp3(yt_path)
        self._delete_file(yt_path)
        return str(converted_path)

    def _download_from_yt(self, song: SongData) -> str:
        """Download a song from a song

        Args:
            song (SearchSongResult): The song to download

        Returns:
            path (str): The path of the downloaded file or None if the download failed
        """
        yt = YouTube(
            f"https://www.youtube.com/watch?v={song.videoId}",
            on_progress_callback=self.download_callback,
        )
        return yt.streams.get_audio_only().download(
            output_path=self.download_path,
            filename=f"{song.title} - {song.get_formatted_artists()}.m4a",
        )

    def _convert_to_mp3(self, path: str) -> str:
        """Convert a m4a file to mp3

        Args:
            path (str): path to the m4a file

        Returns:
            str: path to the mp3 file
        """
        audio = AudioSegment.from_file(path)
        audio.export(path.replace(".m4a", ".mp3"), format="mp3")
        return path.replace(".m4a", ".mp3")

    def _delete_file(self, path: str) -> None:
        """Delete a file

        Args:
            path (str): path to the file
        """
        Path(path).unlink(missing_ok=True)

    def download_callback(self, stream: str, chunk: bytes, total: int) -> None:
        """Callback for the download process

        Args:
            stream (str): The stream of the download
            chunk (bytes): The chunk of the download
            total (int): The total size of the download
        """
        print(f"Downloading {chunk}/{total}")



from glob import glob


def fetch_songs_from_folder(folder_path: str) -> list[str | None]:
    """Fetch all the songs from a folder
    Args:
        folder_path (str): The path of the folder
    Returns:
        list[Song]: The list of songs found
    """
    songs = []
    for file in glob(f"{folder_path}/*.mp3"):
        songs.append(file)
    return songs


def fetch_lyrics_from_folder(folder_path: str) -> list[str | None]:
    """Fetch all the lyrics from a folder
    Args:
        folder_path (str): The path of the folder
    Returns:
        list[str]: The list of lyrics found
    """
    lyrics = []
    for file in glob(f"{folder_path}/*.md"):
        lyrics.append(file)
    return lyrics


if __name__ == "__main__":
    pass



from pathlib import Path
from pymusicterm.api.ytmusic import SearchResult, YTMusic
from pymusicterm.api.player import MusicPlayer
from pymusicterm.api.downloader import Downloader
from pymusicterm.player.media_control import MediaControl
from pymusicterm.setting import SettingManager
from pymusicterm.api.local import fetch_lyrics_from_folder, fetch_songs_from_folder
from random import shuffle
from pymusicterm.api.lyrics import LyricsDownloader

    
    
class PyMusicTermPlayer:
    def __init__(self, setting: SettingManager, media_control: MediaControl) -> None:
        self.media_control = media_control
        self.setting = setting
        self.music_player = MusicPlayer(self.setting.volume)
        lyrics = LyricsDownloader(self.setting.lyrics_dir)
        self.ytm = YTMusic(lyrics)
        self.downloader = Downloader(self.setting.music_dir)
        self.list_of_downloaded_songs: list[str] = fetch_songs_from_folder(
            self.setting.music_dir
        )
        self.dict_of_lyrics: dict[str, str] = self.map_lyrics_to_song()
        self.dict_of_song_result: dict[str, SearchResult] = {}
        self.current_song_index = 0
    

    def query(self, query: str, filter: str) -> list[SearchResult]:
        """Query the YTMusic API for a song

        Args:
            query (str): the query to search for
            filter (str): the filter to use

        Returns:
            list[SearchSongResult | SearchVideoResult]: the list of the results found
        """
        result = self.ytm.search(query, filter)

        self.dict_of_song_result.clear()
        for song in result:
            self.dict_of_song_result[song.videoId] = song
        return result

    def play_from_ytb(self, video_id: str) -> None:
        """Play a song from the list of results

        Args:
            id (int): the index of the song to play
        """
        song = self.dict_of_song_result[video_id]
        path = self.downloader.download(song)
        self.ytm.get_lyrics(song)
        self.list_of_downloaded_songs = fetch_songs_from_folder(self.setting.music_dir)
        self.list_of_lyrics = self.map_lyrics_to_song()
        self.music_player.load_song(str(path))
        self.music_player.play_song()
        self.media_control.on_playback()

    def play_from_list(self, id: int) -> None:
        """Play a song from the list of downloaded songs

        Args:
            id (int): the index of the song to play
        """
        self.current_song_index = id
        self.list_of_downloaded_songs = fetch_songs_from_folder(self.setting.music_dir)
        self.list_of_lyrics = self.map_lyrics_to_song()
        self.music_player.load_song(self.list_of_downloaded_songs[id])
        self.music_player.play_song()
        self.media_control.on_playback()


    def previous(self) -> None:
        """Play the previous song"""
        if self.current_song_index == 0:
            self.current_song_index = len(self.list_of_downloaded_songs) - 1
        else:
            self.current_song_index -= 1
        self.music_player.load_song(self.list_of_downloaded_songs[self.current_song_index])
        self.music_player.play_song()
        self.media_control.on_playback()


    def next(self) -> None:
        """Play the next song"""
        if self.current_song_index == len(self.list_of_downloaded_songs) - 1:
            self.current_song_index = 0
        else:
            self.current_song_index += 1
        self.music_player.load_song(self.list_of_downloaded_songs[self.current_song_index])
        self.music_player.play_song()
        self.media_control.on_playback()

    def seek(self, time: float = 10) -> None:
        """Seek forward or backward

        Args:
            time (float, optional): The time to seek in seconds. Defaults to 10.
        """
        self.music_player.position += time

    def suffle(self) -> None:
        """Shuffle the list of downloaded songs"""
        shuffle(self.list_of_downloaded_songs)

    def loop_at_end(self) -> bool:
        """Loop at the end"""
        self.music_player.loop_at_end = not self.music_player.loop_at_end
        self.setting.loop = self.music_player.loop_at_end
        return self.music_player.loop_at_end

    def update(self) -> None:
        """Update the player"""
        if self.music_player.loop_at_end:
            return
        if self.music_player.position == 0 and not self.music_player.playing:
            self.next()

    def map_lyrics_to_song(self) -> None:
        """Map the lyrics to the songs"""
        list_of_lyrics: dict[str, str] = {}
        for song in fetch_songs_from_folder(self.setting.music_dir):
            lyric = self.setting.lyrics_dir + f"/{Path(song).stem.removesuffix('.mp3')}.md"
            list_of_lyrics[song] = Path(lyric)
        return list_of_lyrics

    def stop(self) -> None:
        self.music_player.unload_song()

    @property
    def song_length(self) -> float:
        """Get the song length"""
        return self.music_player.song_length

    @property
    def position(self) -> float:
        """Get the current position"""
        return self.music_player.position

    def volume(self, value: float) -> None:
        """Get the volume up"""
        self.music_player.volume += value
        self.media_control.on_volume()
        self.setting.volume = self.music_player.volume

    @property
    def playing(self) -> bool:
        """Get the playing status"""
        return self.music_player.playing

    def pause_song(self) -> None:
        """Pause the song"""
        self.music_player.pause_song()
        self.media_control.on_playpause()

    def resume_song(self) -> None:
        """Resume the song"""
        self.music_player.resume_song()
        self.media_control.on_playpause()